"""Wrappers for pysam files."""

from pysam.libcalignmentfile import AlignmentFile as PysamAlignmentFile


class RTAlignmentFile(PysamAlignmentFile):
    """Wrapper for pysam.AlignmentFile to provide filtering on fetch."""

    def __new__(cls, *args, **kwargs):
        """
        Create a wrapper for pysam.AlignmentFile.

        Parameters:
            *args (list): Positional arguments for pysam.FastaFile()
            **kwargs (dict): Keyword arguments for pysam.FastaFile()

        Returns:
            PysamAlignmentFile
        """
        kwargs.pop('min_quality', None)
        kwargs.pop('min_length', None)
        return PysamAlignmentFile.__new__(cls, *args, **kwargs)

    def __init__(self, *args, min_quality=0, min_length=0, **kwargs):
        """
        Create a wrapper for pysam.AlignmentFile.

        Parameters:
            *args (list): Positional arguments for pysam.FastaFile()
            min_quality (int): Minimum read quality
            min_length (int): Minimum read length
            **kwargs (dict): Keyword arguments for pysam.FastaFile()
        """
        PysamAlignmentFile.__init__(self)

        self._checklist = []

        if min_quality > 0:
            self._min_quality = min_quality
            self._checklist.append(self._check_quality)

        if min_length > 0:
            self._min_length = min_length
            self._checklist.append(self._check_length)

    @property
    def exclude_reads(self):
        """
        Names of reads not to be fetched.

        Returns:
            iterable
        """
        return self._exclude_reads

    @exclude_reads.setter
    def exclude_reads(self, read_names):
        """
        Provide a list of read names to be skipped during fetch.

        Parameters:
            read_names (iterable): Reads to skip
        """
        self._exclude_reads = set(read_names)
        self._checklist.append(self._check_read_name)

    def fetch(self, *args, **kwargs):
        """
        Fetch reads aligned in a region.

        Parameters:
            *args (list): Positional arguments for pysam.FastaFile.fetch
            *kwargs (list): Keyword arguments for pysam.FastaFile.fetch

        Yields:
            Reads
        """
        if 'region' in kwargs:
            kwargs['region'] = str(kwargs['region'])  # noqa:WPS529
        try:
            iterator = super().fetch(*args, **kwargs)
        except ValueError:
            return
        for read in iterator:
            if self._check_read(read):
                yield read

    def fetch_by_position(self, *args, **kwargs):
        """
        Retrieve reads that all start at the same point on the reference.

        Parameters:
            *args (list): Positional arguments for fetch
            **kwargs (dict): Named arguments for fetch

        Yields:
            Lists containing reads
        """
        iterator = self.fetch(*args, **kwargs)

        first_read = next(iterator, None)
        if first_read is None:
            return

        reads = [first_read]
        ref_start = first_read.reference_start

        for read in iterator:
            if read.reference_start == ref_start:
                reads.append(read)
            else:
                yield reads
                reads = [read]
                ref_start = read.reference_start
        yield reads

    # 77: NOT_MAPPED
    # 141: NOT_MAPPED
    # 512: QC_FAIL
    # 256: IS_SECONDARY
    # 1024: IS_DUPLICATE
    _flags_to_toss = {77, 141, 512, 256, 1024}
    _paired_flags_to_keep = {99, 147, 83, 163}

    def _check_quality(self, read):
        return read.mapping_quality >= self._min_quality

    def _check_length(self, read):
        return read.query_length >= self._min_length

    def _check_read_name(self, read):
        return read.query_name not in self._exclude_reads

    def _check_read(self, read):
        if read.has_tag('SA'):
            return False
        if read.flag in self._flags_to_toss:
            return False
        if read.is_paired and read.flag not in self._paired_flags_to_keep:
            return False

        for check in self._checklist:
            if not check(read):
                return False
        return True



"""Commandline tool for REDItools."""

import sys

from reditools import analyze, homopolymerics, index, annotate


def usage():
    """Print program usage."""
    print("""usage: reditools {analyze,find-repeats,index,annotate}

REDItools3

Run Modes:
  analyze            Find editing events in one or more alignment files.

  find-repeats       Find repetitive elements in a genome.

  index              Calculate editing indices from the output of `analyze`
                     mode.

  annotate           Annotate REDItools RNA output with DNA output
""")


if __name__ == '__main__':
    if len(sys.argv) > 1:
        command = sys.argv.pop(1)
        match command:
            case 'analyze':
                analyze.main()
            case 'find-repeats':
                homopolymerics.main()
            case 'index':
                index.main()
            case 'annotate':
                annotate.main()
            case _:
                usage()
    else:
        usage()



import argparse
from reditools import file_utils
import csv
import sys


class RTAnnotater:
    def __init__(self, rna_file, dna_file):
        self.rna_file = rna_file
        self.dna_file = dna_file
        self.contig_order = self._load_contig_order()

    def _load_contig_order(self):
        contigs = {}
        idx = 1
        with file_utils.open_stream(self.rna_file, 'r') as stream:
            reader = csv.reader(stream, delimiter='\t')
            last_contig = next(reader)[0]
            contigs[last_contig] = 0
            for row in reader:
                if row[0] == last_contig:
                    continue
                contigs[row[0]] = idx
                idx += 1
                last_contig = row[0]
        return contigs

    def _cmp_position(self, rna_contig, rna_pos, dna_contig, dna_pos):
        rna_contig = self.contig_order[rna_contig]
        dna_contig = self.contig_order.get(dna_contig, len(self.contig_order))
        if rna_contig < dna_contig:
            return -1
        if rna_contig > dna_contig:
            return 1
        rna_pos = int(rna_pos)
        dna_pos = int(dna_pos)
        if rna_pos < dna_pos:
            return -1
        if rna_pos > dna_pos:
            return 1
        return 0

    def _annotate_row(self, rna_row, dna_row):
        rna_row['gCoverage-q30'] = dna_row['Coverage-q30']
        rna_row['gMeanQ'] = dna_row['MeanQ']
        rna_row['gBaseCount[A,C,G,T]'] = dna_row['BaseCount[A,C,G,T]']
        rna_row['gAllSubs'] = dna_row['AllSubs']
        rna_row['gFrequency'] = dna_row['Frequency']
        return rna_row

    def _compare_files(self):
        with file_utils.open_stream(self.rna_file, 'r') as rna_stream, \
                file_utils.open_stream(self.dna_file, 'r') as dna_stream:
            rna_reader = csv.DictReader(rna_stream, delimiter='\t')
            dna_reader = csv.DictReader(dna_stream, delimiter='\t')

            rna_entry = next(rna_reader, None)
            dna_entry = next(dna_reader, None)

            while rna_entry is not None:
                if dna_entry is None:
                    yield rna_entry
                    rna_entry = next(rna_reader, None)
                    continue
                cmp = self._cmp_position(
                    rna_entry['Region'],
                    rna_entry['Position'],
                    dna_entry['Region'],
                    dna_entry['Position'])
                if cmp == 0:
                    yield self._annotate_row(rna_entry, dna_entry)
                    rna_entry = next(rna_reader, None)
                    dna_entry = next(dna_reader, None)
                elif cmp > 0:
                    dna_entry = next(dna_reader, None)
                else:
                    yield rna_entry
                    rna_entry = next(rna_reader, None)

    def annotate(self, stream):
        writer = csv.DictWriter(stream, delimiter='\t', fieldnames=[
            'Region',
            'Position',
            'Reference',
            'Strand',
            'Coverage-q30',
            'MeanQ',
            'BaseCount[A,C,G,T]',
            'AllSubs',
            'Frequency',
            'gCoverage-q30',
            'gMeanQ',
            'gBaseCount[A,C,G,T]',
            'gAllSubs',
            'gFrequency'])
        writer.writeheader()
        writer.writerows(self._compare_files())


def parse_options():
    """
    Parse commandline options for REDItools.

    Returns:
        namespace: commandline args
    """
    parser = argparse.ArgumentParser(
        prog='reditools annotate',
        description='Annotates RNA REDItools output with DNA output.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        'rna_file',
        help='The REDItools output from RNA data',
    )
    parser.add_argument(
        'dna_file',
        help='The REDItools output from corresponding DNA data',
    )
    return parser.parse_args()


def main():
    options = parse_options()
    x = RTAnnotater(options.rna_file, options.dna_file)
    x.annotate(sys.stdout)



"""Commandline tool for REDItools."""

import argparse
import csv
import sys
import traceback
from multiprocessing import Process, Queue
from queue import Empty as EmptyQueueException
from tempfile import NamedTemporaryFile

from reditools import file_utils, reditools, utils
from reditools.alignment_manager import AlignmentManager
from reditools.logger import Logger
from reditools.region import Region

_contig = 'contig'
_start = 'start'
_stop = 'stop'

fieldnames = [
    'Region',
    'Position',
    'Reference',
    'Strand',
    'Coverage-q30',
    'MeanQ',
    'BaseCount[A,C,G,T]',
    'AllSubs',
    'Frequency',
    'gCoverage-q30',
    'gMeanQ',
    'gBaseCount[A,C,G,T]',
    'gAllSubs',
    'gFrequency',
]


def setup_alignment_manager(options):
    """
    Create an AlignmentManager for REDItools.

    Parameters:
        options (namespace): Commandline arguments

    Returns:
        AlignmentManager
    """
    sam_manager = AlignmentManager(
        ignore_truncation=True,
    )
    sam_manager.min_quality = options.min_read_quality
    sam_manager.min_length = options.min_read_length
    for sam in options.file:
        sam_manager.add_file(
            sam,
            options.exclude_reads,
        )
    return sam_manager


def setup_rtools(options):  # noqa:WPS213,WPS231
    """
    Create a REDItools object.

    Parameters:
        options (namespace): Commandline arguments from argparse

    Returns:
        A configured REDItools object
    """
    if options.dna:
        rtools = reditools.REDItoolsDNA()
    else:
        rtools = reditools.REDItools()

    if options.debug:
        rtools.log_level = Logger.debug_level
    elif options.verbose:
        rtools.log_level = Logger.info_level

    if options.load_omopolymeric_file:
        regions = file_utils.read_bed_file(options.load_omopolymeric_file)
        rtools.add_exclude_regions(regions)

    if options.variants:
        rtools.specific_edits = [_.upper() for _ in options.variants]

    if options.bed_file:
        for fname in options.bed_file:
            regions = file_utils.read_bed_file(fname)
            rtools.add_target_regions(regions)
    if options.exclude_regions:
        for fname in options.exclude_regions:
            regions = file_utils.read_bed_file(fname)
            rtools.add_exclude_regions(regions)
    if options.reference:
        rtools.add_reference(options.reference)

    if options.splicing_file:
        rtools.splice_positions = file_utils.load_splicing_file(
            options.splicing_file,
            options.splicing_span,
        )
        rtools.add_exclude_regions(regions)

    rtools.min_base_position = options.min_base_position
    rtools.max_base_position = options.max_base_position
    rtools.min_base_quality = options.min_base_quality

    rtools.min_column_length = options.min_read_depth
    rtools.min_edits = options.min_edits
    rtools.min_edits_per_nucleotide = options.min_edits_per_nucleotide
    rtools.strand = options.strand
    rtools.max_alts = options.max_editing_nucleotides

    rtools.strand_confidence_threshold = options.strand_confidence_threshold

    if options.strand_correction:
        rtools.use_strand_correction()
    if options.exclude_multis:
        rtools.only_one_alt()

    return rtools


def region_args(bam_fname, region, window):
    """
    Split a region into segments for paralllel processing.

    Parameters:
        bam_fname (str): BAM file to collect contig info from
        region (Region): Genomic region to split
        window (int): How large the sub regions should be.

    Returns:
        (list): Sub regions
    """
    if region is not None:
        if window:
            return region.split(window)
        return [region]

    args = []
    for contig, size in utils.get_contigs(bam_fname):
        region = Region(contig=contig, start=1, stop=size+1)
        if window:
            args.extend(region.split(window))
        else:
            args.append(region)
    return args


def write_results(rtools, sam_manager, file_name, region, output_format,
                  temp_dir):
    """
    Write the results from a REDItools analysis to a temporary file.

    Parameters:
        rtools (REDItools): REDItools instance
        sam_manager (AlignmentManager): Source of reads
        file_name (string): Input file name for analysis
        region: Region to analyze
        output_format (dict): keyword arguments for csv.writer constructor.

    Returns:
        string: Name of the temporary file.
    """
    with NamedTemporaryFile(mode='w', delete=False, dir=temp_dir) as stream:
        writer = csv.writer(stream, **output_format)
        for rt_result in rtools.analyze(sam_manager, region):
            variants = rt_result.variants
            writer.writerow([
                rt_result.contig,
                rt_result.position,
                rt_result.reference,
                rt_result.strand,
                rt_result.depth,
                f'{rt_result.mean_quality:.2f}',
                rt_result.per_base_depth,
                ' '.join(sorted(variants)) if variants else '-',
                f'{rt_result.edit_ratio:.2f}',
                '-', '-', '-', '-', '-',
            ])
        return stream.name


def run(options, in_queue, out_queue):
    """
    Analyze a genomic segment using REDItools.

    Parameters:
        options (namesapce): Configuration options from argparse for REDItools
        in_queue (Queue): Queue of input arguments for analysis
        out_queue (Queue): Queue to store paths to analysis results

    Returns:
        bool: True if the in_queue is empty
    """
    try:
        rtools = setup_rtools(options)
        while True:
            args = in_queue.get()
            if args is None:
                return True
            sam_manager = setup_alignment_manager(options)
            idx, region = args
            file_name = write_results(
                rtools,
                sam_manager,
                options.file,
                region,
                options.output_format,
                options.temp_dir,
            )
            out_queue.put((idx, file_name))
    except Exception as exc:
        if options.debug:
            traceback.print_exception(*sys.exc_info())
        sys.stderr.write(f'[ERROR] ({type(exc)}) {exc}\n')
        sys.exit(1)


def parse_options():  # noqa:WPS213
    """
    Parse commandline options for REDItools.

    Returns:
        namespace: commandline args
    """
    parser = argparse.ArgumentParser(
        prog="reditools analyze",
        description='REDItools3',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        'file',
        nargs='+',
        help='The bam file(s) to be analyzed.',
    )
    parser.add_argument(
        '-r',
        '--reference',
        help='Reference FASTA file.',
    )
    parser.add_argument(
        '-o',
        '--output-file',
        help='Path to write output to.',
        default='/dev/stdout',
    )
    parser.add_argument(
        '-s',
        '--strand',
        choices=(0, 1, 2),
        type=int,
        default=0,
        help='Strand: this can be 0 (unstranded),' +
        '1 (second strand oriented) or ' +
        '2 (first strand oriented).',
    )
    parser.add_argument(
        '-a',
        '--append-file',
        action='store_true',
        help='Appends results to file (and creates if not existing).',
    )
    parser.add_argument(
        '-g',
        '--region',
        help='Only analyzes the specified region.',
    )
    parser.add_argument(
        '-m',
        '--load-omopolymeric-file',
        help='BED file of omopolymeric positions.',
    )
    parser.add_argument(
        '-sf',
        '--splicing-file',
        help='The file containing splicing site positions.',
    )
    parser.add_argument(
        '-ss',
        '--splicing-span',
        type=int,
        default=4,
        help='The splicing span.',
    )
    parser.add_argument(
        '-mrl',
        '--min-read-length',
        type=int,
        default=30,  # noqa:WPS432
        help='Reads with length below -mrl will be discarded.',
    )
    parser.add_argument(
        '-q',
        '--min-read-quality',
        type=int,
        default=20,  # noqa:WPS432
        help='Reads with mapping quality below -q will be discarded.',
    )
    parser.add_argument(
        '-bq',
        '--min-base-quality',
        type=int,
        default=30,  # noqa:WPS432
        help='Base quality below -bq will bed discarded.',
    )
    parser.add_argument(
        '-mbp',
        '--min-base-position',
        type=int,
        default=0,
        help='Ignores the first -mbp bases in each read.',
    )
    parser.add_argument(
        '-Mbp',
        '--max-base-position',
        type=int,
        default=0,
        help='Ignores the last -Mpb bases in each read.',
    )
    parser.add_argument(
        '-l',
        '--min-read-depth',
        type=int,
        default=1,
        help='Only report on positions with at least -l read depth',
    )
    parser.add_argument(
        '-e',
        '--exclude-multis',
        default=False,
        help='Do not report any position with more than one alternate base.',
        action='store_true',
    )
    parser.add_argument(
        '-men',
        '--min-edits-per-nucleotide',
        type=int,
        default=0,
        help='Positions with fewer than -men edits will not be discarded.',
    )
    parser.add_argument(
        '-me',
        '--min-edits',
        type=int,
        default=1,
        help='The minimum number of editing events (per position). ' +
        'Positions with fewer than -me edits will be discarded.',
    )
    parser.add_argument(
        '-Men',
        '--max-editing-nucleotides',
        type=int,
        default=4,  # noqa:WPS432
        help='The maximum number of editing nucleotides, from 0 to 3 ' +
        '(per position). Positions whose columns have more than ' +
        '"max-editing-nucleotides" will not be included in the analysis.',
    )
    parser.add_argument(
        '-T',
        '--strand-confidence-threshold',
        type=float,
        default=0.7,  # noqa:WPS432
        help='Only report the strandedness if at least -T proportion of ' +
        'reads are of a given strand.',
    )
    parser.add_argument(
        '-C',
        '--strand-correction',
        default=False,
        help='Strand correction. Once the strand has been inferred, ' +
        'only bases according to this strand will be selected.',
        action='store_true',
    )
    parser.add_argument(
        '-V',
        '--verbose',
        default=False,
        help='Run in verbose mode.',
        action='store_true',
    )
    parser.add_argument(
        '-N',
        '--dna',
        default=False,
        help='Run REDItools on DNA-Seq data.',
        action='store_true',
    )
    parser.add_argument(
        '-B',
        '--bed_file',
        nargs='+',
        help='Only analyze regions in the provided BED file.',
    )
    parser.add_argument(
        '-t',
        '--threads',
        help='Number of threads for parallel processing.',
        type=int,
        default=1,
    )
    parser.add_argument(
        '--temp-dir',
        help='Location to save temporary files')
    parser.add_argument(
        '-w',
        '--window',
        help='How many bp should be processed by each thread at a time. ' +
        'Zero uses the full contig.',
        type=int,
        default=0,
    )
    parser.add_argument(
        '-k',
        '--exclude_regions',
        nargs='+',
        help='Skip regions in the provided BED file(s).',
    )
    parser.add_argument(
        '-E',
        '--exclude_reads',
        help='Text file listing read names to exclude from analysis.',
    )
    parser.add_argument(
        '-d',
        '--debug',
        default=False,
        help='Run in debug mode.',
        action='store_true',
    )
    parser.add_argument(
        '-v',
        '--variants',
        nargs='*',
        default=['all'],
        help='Which editing events to report. Edits should be two characters, '
        'separated by spaces. Use "all" to report all variants.',
    )

    return parser.parse_args()


def check_dead(processes):
    """
    Look through processes to determine if any have died unexpectedly.

    If any process has an exit code of 1, this method will terminate all other
    processes and then exit with code 1.

    Parameters:
        processes (list): Processes to check
    """
    for proc in processes:
        if proc.exitcode == 1:
            for to_kill in processes:
                to_kill.kill()
            sys.stderr.write('[ERROR] Killing job\n')
            sys.exit(1)


def main():
    """Perform RNA editing analysis."""
    options = parse_options()
    options.output_format = {'delimiter': '\t', 'lineterminator': '\n'}
    options.encoding = 'utf-8'
    if options.exclude_reads:
        options.exclude_reads = file_utils.load_text_file(
            options.exclude_reads,
        )

    # Put analysis chunks into queue
    regions = region_args(
        options.file[0],
        Region(string=options.region) if options.region else None,
        window=options.window,
    )

    in_queue = Queue()
    for args in enumerate(regions):
        in_queue.put(args)
    for _ in range(options.threads):
        in_queue.put(None)

    # Start parallel jobs
    out_queue = Queue()
    processes = [
        Process(
            target=run,
            args=(options, in_queue, out_queue),
        ) for _ in range(options.threads)
    ]
    concat_output(
        options,
        monitor(processes, out_queue, in_queue.qsize()),
    )


def monitor(processes, out_queue, chunks):
    """
    Monitor parallel REDItools jobs.

    Parameters:
        processes (list): Threads
        out_queue (Queue): Output of threads
        chunks (int): Number of chunks for analysis

    Returns:
        list: Temporary files containing the output of each chunk.
    """
    tfs = [None for _ in range(chunks - len(processes))]

    for prc in processes:
        prc.start()

    while None in tfs:
        try:
            idx, fname = out_queue.get(block=False, timeout=1)
            tfs[idx] = fname
        except EmptyQueueException:
            check_dead(processes)
    return tfs


def concat_output(options, tfs):
    """
    Write the output of a REDItools analysis.

    Parameters:
        options (namespace): Commandline options for file formatting.
        tfs (list): Temporary files containing REDItools results
    """
    # Setup final output file
    if options.output_file:
        mode = 'a' if options.append_file else 'w'
        stream = file_utils.open_stream(
            options.output_file,
            mode,
            encoding=options.encoding,
        )
    else:
        stream = sys.stdout

    with stream:
        writer = csv.writer(stream, **options.output_format)
        if not options.append_file:
            writer.writerow(fieldnames)
        file_utils.concat(stream, *tfs, encoding=options.encoding)



"""Wrappers for PysamFastaFile."""

from pysam.libcfaidx import FastaFile as PysamFastaFile


class RTFastaFile(PysamFastaFile):
    """Wrapper for pysam.FastaFile to provide sequence cache."""

    def __new__(cls, *args, **kwargs):
        r"""
        Create a wrapper for pysam.FastaFile.

        Parameters:
            *args (list): positional arguments for PysamFastaFile constructor
            **kwargs (dict): named arguments for PysamFastaFile constructor

        Returns:
            PysamFastaFIle
        """
        return PysamFastaFile.__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        r"""
        Create a wrapper for pysam.FastaFile.

        Parameters:
            *args (list): positional arguments for PysamFastaFile constructor
            **kwargs (dict): named arguments for PysamFastaFile constructor
        """
        PysamFastaFile.__init__(self)

        self._contig_name = False
        self._contig_cache = None

    def get_base(self, contig, *position):
        """
        Retrieve the base at the given position.

        Parameters:
            contig (string): Chromsome name
            position (int): Zero-indexed position on reference

        Returns:
            Base the position as a string.

        Raises:
            IndexError: The position is not within the contig
        """
        if contig != self._contig_name:
            self._update_contig_cache(contig)
        try:
            return [self._contig_cache[idx] for idx in position]
        except IndexError as exc:
            raise IndexError(
                f'Base position {position} is outside the bounds of ' +
                '{contig}. Are you using the correct reference?',
            ) from exc

    def _update_contig_cache(self, contig):
        keys = (contig, f'chr{contig}', contig.replace('chr', ''))
        for ref in keys:
            if ref in self:
                self._contig_cache = self.fetch(ref).upper()
                self._contig_name = contig
                return
        raise KeyError(f'Reference name {contig} not found in FASTA file.')



"""Miscellaneous utility functions."""

import csv
import os
from gzip import open as gzip_open

from reditools.region import Region


def open_stream(path, mode='rt', encoding='utf-8'):
    """
    Open a input or output stream from a file, accounting for gzip.

    Parameters:
        path (str): Path to file for reading or writing
        mode (str): File mode
        encoding (str): File encoding

    Returns:
        TextIOWrapper to the file
    """
    if path.endswith('gz'):
        return gzip_open(path, mode, encoding=encoding)
    return open(path, mode, encoding=encoding)  # noqa:WPS515


def read_bed_file(path):
    """
    Return an iterator for a BED file.

    Parameters:
        path (str): Path to a BED file for reading.

    Yields:
        BED file contents as Regions.
    """
    stream = open_stream(path)
    reader = csv.reader(
        filter(lambda row: row[0] != '#', stream),
        delimiter='\t',
    )
    yield from (Region(
        contig=row[0],
        start=row[1],
        stop=row[2],
        ) for row in reader
    )


def concat(output, *fnames, clean_up=True, encoding='utf-8'):
    """
    Combine one or more files into another file.

    Parameters:
        output (file): A file like object for writing
        *fnames (string): Paths to files for concatenation
        clean_up (bool): If True, deletes the files after concatenation
        encoding (string): File encoding
    """
    for fname in fnames:
        with open(fname, 'r', encoding=encoding) as stream:
            for line in stream:
                output.write(line)
        if clean_up:
            os.remove(fname)


def load_splicing_file(splicing_file, splicing_span):
    """
    Read splicing positions from a file.

    Parameters:
        splicing_file (str): File path
        splicing_span(int): Width of splice sites

    Yeilds:
        Splicing file contents as Regions.
    """
    strand_map = {'-': 'D', '+': 'A'}

    stream = open_stream(splicing_file)
    reader = csv.reader(
        filter(lambda row: row[0] != '#', stream),
        delimiter=' ',
    )
    for row in reader:
        contig = row[0]
        span = int(row[1])
        splice = row[3]
        strand = row[4]

        coe = -1 if strand_map.get(strand, None) == splice else 1
        start = 1 + span
        stop = start + splicing_span * coe
        if start > stop:
            start, stop = stop, start
        yield Region(contig=contig, start=start, stop=stop)


def load_text_file(file_name):
    """
    Extract file contents to a list.

    Parameters:
        file_name (str): The file to open.

    Returns:
        List of content
    """
    with open_stream(file_name, 'r') as stream:
        return [line.strip() for line in stream]



"""Commandline tool for REDItools."""

import argparse
import csv
import sys
from itertools import permutations
from json import loads as load_json

from reditools.file_utils import open_stream, read_bed_file
from reditools.region import Region

_ref = 'Reference'
_position = 'Position'
_contig = 'Region'
_count = 'BaseCount[A,C,G,T]'
_strand = 'Strand'
_nucs = 'ACGT'
_ref_set = {f'{nuc}-{nuc}' for nuc in _nucs}


class Index(object):
    """Utility for calculating editing indices."""

    def __init__(self, region=None, strand=0):
        """
        Create a new Index.

        Parameters:
            region (Region): Limit results to the given genomic region
            strand (int): Either 0, 1, or 2 for unstranded, reverse, or forward
        """
        self.targets = {}
        self.exclusions = {}
        self.counts = {'-'.join(_): 0 for _ in permutations(_nucs, 2)}
        self.region = region
        self.strand = ['*', '-', '+'][strand]

    def add_target_from_bed(self, fname):
        """
        Only report index data for regions from a given bed file.

        Parameters:
            fname (str): Path to BED formatted file.
        """
        for region in read_bed_file(fname):
            self.targets[region.contig] = update_region_dict(
                self.targets,
                region,
            )

    def add_exclusions_from_bed(self, fname):
        """
        Exclude index data for regions from a given bed file.

        Parameters:
            fname (str): Path to BED formatted file.
        """
        for region in read_bed_file(fname):
            self.exclusions[region.contig] = update_region_dict(
                self.exclusions,
                region,
            )

    def in_region_list(self, region_list, contig, position):
        """
        Check if a genomic position is in a list of regions.

        Parameters:
            region_list (dict): Region list to check
            contig (str): Contig/Chromsome name
            position (int): Coordinate

        Returns:
            True if the position is present, else False
        """
        return position in region_list.get(contig, [])

    def in_targets(self, contig, position):
        """
        Check if a genomic position is in the target list.

        Parameters:
            contig (str): Contig/Chromsome name
            position (int): Coordiante

        Returns:
            True if there are no targets or the position is in the target
            list; else False
        """
        return not self.targets or self.in_region_list(self.targets)

    def in_exclusions(self, contig, position):
        """
        Check if a genomic position is in the exclusions list.

        Parameters:
            contig (str): Contig/Chromsome name
            position (int): Coordiante

        Returns:
            True if there are no exclusions or the position is in the
            exclusions list; else False
        """
        return self.exclusions and self.in_region_list(self.exclusions)

    def do_ignore(self, row):
        """
        Check whether a row should meets analysis criteria.

        Parameters:
            row (dict): Row from REIDtools output file.

        Returns:
            True if the row should be discarded; else False
        """
        if '*' != self.strand != row[_strand]:
            return True
        if self.region:
            if not self.region.contains(row[_contig], row[_position]):
                return True
        if self.in_exclusions(row[_contig], row[_position]):
            return True
        return not self.in_targets(row[_contig], row[_position])

    def add_rt_output(self, fname):
        """
        Count the number of reads with matches and substitutions.

        Parameters:
            fname (str): File path to a REDItools output
        """
        stream = open_stream(fname)
        reader = csv.DictReader(stream, delimiter='\t')
        for row in reader:
            if self.do_ignore(row):
                continue
            ref = row[_ref]
            reads = load_json(row[_count])
            for nuc, count in zip(_nucs, reads):
                key = f'{nuc}-{ref}'
                self.counts[key] = self.counts.get(key, 0) + count
        stream.close()

    def calc_index(self):
        """
        Compute all editing indices.

        Returns:
            Dictionary of indices
        """
        keys = set(self.counts) - _ref_set
        indices = {}
        for idx in keys:
            ref = idx[-1]
            numerator = self.counts[idx]
            denominator = self.counts.get(self.ref_edit(ref), 0) + numerator
            if denominator == 0:
                indices[idx] = 0
            else:
                indices[idx] = numerator / denominator
        return indices

    def ref_edit(self, ref):
        """
        Format a base as a non-edit.

        Parameters:
            ref (str): Reference base

        Returns:
            A string in the format of {ref}-{ref}
        """
        return f'{ref}-{ref}'


def parse_options():  # noqa:WPS213
    """
    Parse commandline options for REDItools.

    Returns:
        namespace: commandline args
    """
    parser = argparse.ArgumentParser(
        prog="reditools index",
        description='REDItools3',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        'file',
        nargs='+',
        help='The REDItools output file to be analyzed',
    )
    parser.add_argument(
        '-o',
        '--output-file',
        default='/dev/stdout',
        help='The output statistics file',
    )
    parser.add_argument(
        '-s',
        '--strand',
        choices=(0, 1, 2),
        type=int,
        default=0,
        help='Strand: this can be 0 (unstranded),' +
        '1 (secondstrand oriented) or ' +
        '2 (firststrand oriented)',
    )
    parser.add_argument(
        '-g',
        '--region',
        help='The genomic region to be analyzed',
    )
    parser.add_argument(
        '-B',
        '--bed_file',
        nargs='+',
        help='Path of BED file containing target regions',
    )
    parser.add_argument(
        '-k',
        '--exclude_regions',
        nargs='+',
        help='Path of BED file containing regions to exclude from analysis',
    )

    return parser.parse_args()


def main():
    """Perform RNA editing analysis."""
    options = parse_options()
    if options.region:
        indexer = Index(Region(string=options.region), strand=options.strand)
    else:
        indexer = Index(strand=options.strand)

    if options.exclude_regions:
        for exc_fname in options.exclude_regions:
            indexer.add_exclusions_from_bed(exc_fname)

    if options.bed_file:
        for trg_fname in options.bed_file:
            indexer.add_target_from_bed(trg_fname)

    if options.output_file:
        stream = open_stream(options.output_file, 'w')
    else:
        stream = sys.stdout

    for fname in options.file:
        indexer.add_rt_output(fname)

    for nuc, idx in sorted(indexer.calc_index().items()):
        stream.write(f'{nuc}\t{idx}\n')


def update_region_dict(region_dict, region):
    """
    Add a region to a region dictionary.

    Parameters:
        region_dict (dict): Region dictionary
        region (Region): Region to add

    Returns:
        An updated copy of region_dict
    """
    return region_dict.get(region.contig, set()) | region.enumerate()


if __name__ == '__main__':
    main()



"""Wrappers for pysam files."""
from itertools import chain

from reditools.alignment_file import RTAlignmentFile


class ReadGroupIter(object):
    """Manages multiple fetch iterators."""

    _iter_idx = 0
    _reads_idx = 1
    _start_idx = 2

    def __init__(self, fetch_iters):
        """
        Combine multiple fetch iterators.

        Parameters:
            fetch_iters (iterable): The iterators to combine.
        """
        self._read_groups = []
        for itr in fetch_iters:
            reads = next(itr, None)
            if reads is None:
                continue
            start = reads[0].reference_start
            self._read_groups.append({
                self._iter_idx: itr,
                self._reads_idx: reads,
                self._start_idx: start,
            })

    def is_empty(self):
        """
        Check if there are still reads left.

        Returns:
            bool: True if empty, else False
        """
        return not self._read_groups

    def next(self):
        """
        Retrieve a list of reads that all start at the same position.

        Returns:
            list: Reads
        """
        position = self._find_start()
        reads = []
        for idx in range(len(self._read_groups) - 1, -1, -1):
            group = self._read_groups[idx]
            if group[self._start_idx] == position:
                reads.append(group[self._reads_idx])
                next_reads = next(group[self._iter_idx], None)
                if next_reads is None:
                    self._read_groups.pop(idx)
                else:
                    self._read_groups[idx] = {
                        self._iter_idx: group[self._iter_idx],
                        self._reads_idx: next_reads,
                        self._start_idx: next_reads[0].reference_start,
                    }
        return reads

    def _find_start(self):
        return min(group[self._start_idx] for group in self._read_groups)


class AlignmentManager(object):
    """
    Manage multiple RTAlignmentFiles with a single fetch.

    Attributes:
        min_quality (int): Minimum read quality (applied during add_file)
        min_length (int): Minimum read length (applied during add_file)
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new manager.

        Parameters:
            *args (list): positional arguments for PysamFastaFile
                constructor
            **kwargs (dict): named arguments for PysamFastaFile
                constructor
        """
        self._bam_args = args
        self._bam_kwargs = kwargs
        self._bams = []
        self.min_quality = 0
        self.min_length = 0
        self.file_list = []

    def add_file(self, fname, exclude_reads=None):
        """
        Add an alignment file to the manager for analysis.

        Parameters:
            fname (str): Path to BAM file
            exclude_reads (set): Read names not to skip
        """
        new_file = RTAlignmentFile(
            fname,
            *self._bam_args,
            min_quality=self.min_quality,
            min_length=self.min_length,
            **self._bam_kwargs,
        )
        new_file.check_index()
        if exclude_reads:
            new_file.exclude_reads = exclude_reads
        self._bams.append(new_file)
        self.file_list.append(fname)

    def fetch_by_position(self, *args, **kwargs):
        """
        Perform combine fetch_by_position for all managed files.

        Parameters:
            *args (list): Positional arguments for
                RTAlignmentFile.fetch_by_position
            **kwargs (dict): Named arguments for
                RTAlignmentFile.fetch_by_position

        Yields:
            list: reads from all managed files that begin at the same position.
        """
        iters = [bam.fetch_by_position(*args, **kwargs) for bam in self._bams]
        rgi = ReadGroupIter(iters)
        while not rgi.is_empty():
            reads = list(chain(*rgi.next()))
            self.position = reads[0].reference_start
            self.contig = reads[0].reference_name
            yield reads



"""Genomic Region Collection."""


class RegionCollection(object):
    """Collections of REDItools3 region objects. This class is meant to
    provide fast lookups of overlaps, and so behaves as a queue."""

    def __init__(self):
        """
        Creates a new region collection.

        """

        self._regions = {}
        self._last_index = None
        self._last_contig = None
        self._sorted = False

    def _sort(self):
        for contig, regions in self._regions.items():
            self._regions[contig] = sorted(
                regions,
                key=lambda _: (_.start, _.stop),
            )
        self._sorted = True

    def contains(self, contig, position):
        """
        Checks whether the given position or range overlaps with the
        collection.

        Parameters:
            contig (str): Chromomsome/contig name from reference.
            position_start (int): Position to check for.

        Returns:
            True if there is an overlap, False otherwise.
        """
        if not self._sorted:
            self._sort()

        if contig != self._last_contig:
            self._last_contig = contig
            self._last_index = 0

        for i in range(self._last_index, len(self._regions.get(contig, []))):
            self._last_index = i
            region = self._regions[contig][i]
            if position < region.start:
                return False
            if position >= region.start and \
                    (region.stop is None or position < region.stop):
                return True
        self._last_index += 1

        return False

    def add_region(self, region):
        """
        Add a region to the collection.

        Parameters:
            region (Region): region to add.
        """
        self._sorted = False
        if region.contig not in self._regions:
            self._regions[region.contig] = []
        self._regions[region.contig].append(region)

    def add_regions(self, regions):
        """
        Add a list or iterable of regions to the collection.

        Parameters:
            regions (iterable): List of regions.
        """
        for r in regions:
            self.add_region(r)



"""Genomic Region."""

import re


class Region(object):
    """Genomic Region."""

    def __init__(self, **kwargs):
        """
        Create a new genomic region.

        Parameters:
            **kwargs (dict):
                string (str): String representation of a region
                OR
                contig (str): Contig name
                start (int): Genomic start
                stop (int): Genomic stop

        Raises:
            ValueError: The contig is missing
        """
        if 'string' in kwargs:
            region = self._parse_string(kwargs['string'])  # noqa:WPS529
            self.contig = region[0]
            self.start = region[1]
            self.stop = region[2]
        else:
            if 'contig' not in kwargs:
                raise ValueError('Region constructor requires a contig.')
            self.contig = kwargs['contig']
            self.start = self._to_int(kwargs.get('start', 1)) - 1
            if 'stop' in kwargs:
                self.stop = self._to_int(kwargs['stop']) - 1
            else:
                self.stop = None

    def __str__(self):
        """
        Put the region into standard string format.

        Returns:
            (str): contig:start-stop
        """
        if self.start > 0:
            if self.stop:
                return f'{self.contig}:{self.start}-{self.stop + 1}'
            return f'{self.contig}:{self.start}'
        return self.contig

    def split(self, window):
        """
        Split the region into a list of smaller regions.

        Parameters:
            window (int): The size of the sub regions in bp

        Returns:
            list

        Raises:
            IndexError: The region is missing a start or stop
        """
        if not self.stop or not self.start:
            raise IndexError('Can only split a region with a start and stop.')
        length = self.stop - self.start
        sub_regions = []
        for offset in range(0, length + 1, window):
            sub_regions.append(Region(
                contig=self.contig,
                start=self.start + offset,
                stop=self.start + offset + window,
            ))
        if self.start < length:
            sub_regions.append(Region(
                contig=self.contig,
                start=sub_regions[-1].stop,
                stop=self.stop,
            ))
        return sub_regions

    def enumerate(self):
        """
        Convert a list of regions into a list of individual positions.

        Returns:
            Set enumerating the individual positions.
        """
        return set(range(self.start, self.stop))

    def contains(self, contig, position):
        """
        Determines if a given genomic location is within the region.

        Parameters:
            contig (str): Contig/Chromosome name
            position (int): Position

        Returns:
            bool
        """
        if self.contig != contig:
            return False
        left = self.start is None or self.start <= position
        right = self.stop is None or position < self.stop
        return left and right

    def _parse_string(self, region_str):
        if region_str is None:
            return None
        region = re.split('[:-]', region_str)
        if not region:
            return None
        contig = region[0]
        start = 0
        stop = None

        if len(region) > 3:
            raise ValueError(f'Unrecognized format: {region_str}.')
        if len(region) > 1:
            start = self._to_int(region[1])
            if len(region) == 3:
                stop = self._to_int(region[2])
        return (contig, start, stop)

    def _to_int(self, number):
        if isinstance(number, str):
            return int(re.sub(r'[\s,]', '', number))
        if number is None:
            return None
        return int(number)



"""Organizational structure for tracking base coverage of genomic positions."""

from reditools.compiled_position import CompiledPosition

inf = float('inf')


class CompiledReads(object):
    """Manager for CompiledPositions."""

    _strands = ('-', '+', '*')

    def __init__(
        self,
        strand=0,
        min_base_position=0,
        max_base_position=inf,
        min_base_quality=0,
    ):
        """
        Create a new CompiledReads object.

        Parameters:
            strand (int): Strand detection mode
            min_base_position (int): Left trims bases
            max_base_position (int): Right trims bases
            min_base_quality (int): Minimum base quality to report
        """
        self._nucleotides = {}
        if strand == 0:
            self.get_strand = lambda _: 2
        elif strand == 1:
            self.get_strand = self._get_strand_one
        else:
            self.get_strand = self._get_strand_two

        self._ref = None
        self._ref_seq = self._get_ref_from_read

        self._qc = {
            'min_base_quality': min_base_quality,
            'min_base_position': min_base_position,
            'max_base_position': max_base_position,
        }

    def add_reference(self, ref):
        """
        Add a reference FASTA file to use.

        Parameters:
            ref (RTFastaFile): Reference sequence
        """
        self._ref = ref
        self._ref_seq = self._get_ref_from_fasta

    def add_reads(self, reads):
        """
        Add iterable of pysam reads to the object.

        The reads are broken down. into individual nucleotides that are
        tracked by chromosomal location.

        Parameters:
            reads (iterable): pysam reads
        """
        for read in reads:
            strand = self._strands[self.get_strand(read)]
            for pos, base, quality, ref in self._prep_read(read):
                try:
                    self._nucleotides[pos].add_base(quality, strand, base)
                except KeyError:
                    self._nucleotides[pos] = CompiledPosition(
                        ref=ref,
                        position=pos,
                        contig=read.reference_name,
                    )
                    self._nucleotides[pos].add_base(quality, strand, base)

    def pop(self, position):
        """
        Remove and return the CompiledPosition at position.

        Method returns None if the position is empty.

        Parameters:
            position (int): The chromosomal location to pop

        Returns:
            A CompiledPosition or None if position is empty.
        """
        return self._nucleotides.pop(position, None)

    def is_empty(self):
        """
        Determine if there are any CompiledPositions still in the object.

        Returns:
            True if the object is empty, else False
        """
        return not self._nucleotides

    def _get_ref_from_read(self, read):
        return list(read.get_reference_sequence().upper())

    def _get_ref_from_fasta(self, read):
        pairs = read.get_aligned_pairs(matches_only=True)
        indices = [ref for _, ref in pairs]
        return self._ref.get_base(read.reference_name, *indices)

    def _qc_base_position(self, read, position):
        return read.query_length - position >= self._qc['max_base_position']

    def _prep_read(self, read):
        pairs = read.get_aligned_pairs(matches_only=True)
        seq = read.query_sequence.upper()
        qualities = read.query_qualities
        ref_seq = self._ref_seq(read)
        while pairs and pairs[0][0] < self._qc['min_base_position']:
            pairs.pop(0)
            ref_seq.pop(0)
        if not pairs:
            return

        while pairs and self._qc_base_position(read, pairs[0][0]):
            offset, ref_pos = pairs.pop(0)
            ref_base = ref_seq.pop(0)
            if ref_base != 'N' != seq[offset]:
                if qualities[offset] >= self._qc['min_base_quality']:
                    yield (ref_pos, seq[offset], qualities[offset], ref_base)

    def _get_strand_one(self, read):
        return read.is_read1 and not read.is_reverse or \
            read.is_read2 and read.is_reverse

    def _get_strand_two(self, read):
        return read.is_read1 and read.is_reverse or \
            read.is_read2 and not read.is_reverse



"""Miscellaneous utility functions."""

import csv
import os
import re
import socket
from collections import defaultdict

from pysam.libcalignmentfile import AlignmentFile
from sortedcontainers import SortedSet

from reditools.file_utils import open_stream


def read_bed_file(path):
    """
    Return an iterator for a BED file.

    Parameters:
        path (str): Path to a BED file for reading.

    Returns:
        Iterator of BED file contents.
    """
    stream = open_stream(path)
    return csv.reader(stream, delimiter='\t')


def enumerate_positions(regions):
    """
    Convert a list of regions into a list of individual positions.

    Parameters:
        regions (list): A list of iterables. Each element must start
                        with a contig and start position. End position
                        is optional. Additional values will be ignored.

    Returns:
        SortedSet enumerating the individual positions.
    """
    positions = defaultdict(SortedSet)
    for region in regions:
        positions[region.contig] |= region.enumerate()
    return positions


def get_hostname_string():
    """
    Retrieve the machine hostname, ip, and proccess ID.

    Returns:
        String in the format "hostname|ip|pid"
    """
    hostname = socket.gethostname()
    ip_addr = socket.gethostbyname(hostname)
    pid = os.getpid()
    return f'{hostname}|{ip_addr}|{pid}'


def check_list(functions, **kwargs):
    """
    Run through a list of functions, determining if any return False.

    Parameters:
        functions (list): A list of function references
        **kwargs: Any arguments to be passed to the members of functions

    Returns:
        False if any function in check_list returns False, else True
    """
    for check in functions:
        if not check(**kwargs):
            return False
    return True


def to_int(string):
    """
    Convert a (potentially formatted) string to an int.

    Parameters:
        string (str): A string representation of an integer

    Returns:
        The integer values of the string.
    """
    return int(re.sub(r'[\s,]', '', string))


def get_contigs(sam_path):
    """
    Retrieve contig or chromsome data from an alignment file.

    Parameters:
        sam_path (string): Path to an alignment file.

    Returns:
        tuple of lists containing the reference names and reference lengths in
        corresponding order
    """
    with AlignmentFile(sam_path, ignore_truncation=True) as sam:
        contigs = list(sam.references)
        sizes = list(sam.lengths)
        indices = range(len(contigs))
        indices = sorted(indices, key=lambda idx: contigs[idx])
        return ((contigs[idx], sizes[idx]) for idx in indices)



"""Quality control for REDItools analyses."""

from reditools import utils
from reditools.logger import Logger


class RTChecks(object):
    """Quality control for REDItools analyses."""

    def __init__(self):
        """Create a RTChecks object."""
        self.check_list = [self.check_is_none]

    def add(self, function):
        """
        Add a QC check.

        Parameters:
            function (RTChecks method): The check to perform
        """
        self.check_list.append(function)

    def discard(self, function):
        """
        Remove a QC check.

        Parameters:
            function (RTChecks method): The check to discard
        """
        if function in self.check_list:
            self.check_list.remove(function)

    def check(self, rtools, bases):
        """
        Perform QC.

        Parameters:
            rtools (REDItools): Object performing analysis
            bases (CompiledPosition): Base position under analysis

        Returns:
            (bool): True of all checks pass, else false
        """
        return utils.check_list(
            self.check_list,
            bases=bases,
            rtools=rtools,
        )

    def check_column_min_length(self, rtools, bases):
        """
        Check read depth.

        Parameters:
            rtools (REDItools): Object performing analysis
            bases (CompiledPosition): Base position under analysis

        Returns:
            (bool): True if the read depth is sufficient
        """
        if len(bases) < rtools.min_column_length:
            rtools.log(
                Logger.debug_level,
                'DISCARDING COLUMN {} [MIN_COLUMN_LEGNTH={}]',
                len(bases),
                rtools.min_column_length,
            )
            return False
        return True

    # Really shouldn't use this one. I have to compute mean_q anyway
    def check_column_quality(self, rtools, bases):
        """
        Check mean quality of the position.

        Parameters:
            rtools (REDItools): Object performing analysis
            bases (CompiledPosition): Base position under analysis

        Returns:
            (bool): True if quality is sufficient
        """
        if bases:
            mean_q = sum(bases.qualities) / len(bases)
        else:
            mean_q = 0
        if mean_q < rtools.min_read_quality:
            rtools.log(
                Logger.debug_level,
                'DISCARD COLUMN mean_quality={} < {}',
                mean_q,
                rtools.min_read_quality,
            )
            return False
        return True

    def check_column_edit_frequency(self, rtools, bases):
        """
        Check the number of edits at the site.

        Parameters:
            rtools (REDItools): Object performing analysis
            bases (CompiledPosition): Base position under analysis

        Returns:
            (bool): True if there are sufficient edits.
        """
        edits_no = len(bases) - bases['REF']
        if edits_no < rtools.min_edits:
            rtools.log(
                Logger.debug_level,
                'DISCARDING COLUMN edits={} < {}',
                edits_no,
                rtools.min_edits,
            )
            return False
        return True

    def check_column_min_edits(self, rtools, bases):
        """
        Check that there are sufficient edit events for each base.

        Parameters:
            rtools (REDItools): Object performing analysis
            bases (CompiledPosition): Base position under analysis

        Returns:
            (bool): True if there are sufficient edits
        """
        for base in "ATCG":
            if base == bases.ref:
                continue
            if bases[base] < rtools.min_edits_per_nucleotide:
                rtools.log(
                    Logger.debug_level,
                    'DISCARDING COLUMN edits={} < {}',
                    bases[base],
                    rtools.min_edits_per_nucleotide,
                )
                return False
        return True

    def check_multiple_alts(self, bases, rtools):
        """
        Check that there is, at most, one alternate base.

        Parameters:
            bases (CompiledPosition): Base position under analysis
            rtools (REDItools): Object running the analysis

        Returns:
            (bool): True if there is zero or one alt
        """
        alts = bases.get_variants()
        if len(alts) < 2:
            rtools.log(
                Logger.debug_level,
                'DISCARD COLUMN alts={} > 1',
                len(alts),
            )
            return False
        return True

    def check_is_none(self, bases, rtools):
        """
        Check if the bases object is None.

        Parameters:
            bases (CompiledPosition): Data for analysis
            rtools (REDItools): Object running the analysis

        Returns:
            (bool): True if bases is not None
        """
        if bases is None:
            rtools.log(Logger.debug_level, 'DISCARD COLUMN no reads')
            return False
        return True

    def check_target_positions(self, bases, rtools):
        """
        Check if the bases object is in a target region.

        Parameters:
            bases (CompiledPosition): Data for analysis
            rtools (REDItools): Object running the analysis

        Returns:
            (bool): True if the position is in a target region
        """
        in_targets = rtools.target_regions.contains(
            bases.contig,
            bases.position,
        )
        if not in_targets:
            rtools.log(
                Logger.debug_level,
                'DISCARD COLUMN not in target regions',
            )
        return in_targets

    def check_exclusions(self, bases, rtools):
        """
        Check if the bases object is in an excluded position.

        Parameters:
            bases (CompiledPosition): Data for analysis
            rtools (REDItools): Object running the analysis

        Returns:
            (bool): True if the position is not excluded
        """
        in_exclusions = rtools.exclude_regions.contains(
            bases.contig,
            bases.position,
        )
        if in_exclusions:
            rtools.log(Logger.debug_level, 'DISCARD COLUMN in excluded region')
            return False
        return True

    def check_max_alts(self, bases, rtools):
        """
        Check that there are no more than a max number of alts.

        Parameters:
            bases (CompiledPosition): Base position under analysis
            rtools (REDItools): Object running the analysis

        Returns:
            (bool): True if there are n or fewer alts
        """

        alts = bases.get_variants()
        if len(alts) > rtools.max_alts:
            rtools.log(
                Logger.debug_level,
                'DISCARD COLUMN alts={} > {}',
                len(alts),
                rtools.max_alts,
            )
            return False
        return True



"""Organizational structure for tracking base coverage of genomic positions."""


class CompiledPosition(object):
    """Tracks base frequency for a genomic position."""

    _bases = 'ACGT'
    _comp = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}

    def __init__(self, ref, contig, position):
        """
        Create a new compiled position.

        Parameters:
            ref (string): The reference base for this position
            contig (string): Chromosome name
            position (int): Genomic coordinate
        """
        self.qualities = []
        self.strands = []
        self.bases = []
        self.counter = False
        self.ref = ref
        self.contig = contig
        self.position = position

    def __len__(self):
        """
        Position depth.

        Returns:
            int
        """
        return len(self.bases)

    def __getitem__(self, base):
        """
        Frequency of a given nucleotide at this position.

        Parameters:
            base (str): The nucleotide (A, C, G, T, or REF)

        Returns:
            int: The total number of reads with the given base
        """
        if not self.counter:
            self.counter = {base: 0 for base in self._bases}
            for base_member in self.bases:
                self.counter[base_member] += 1
        if base.upper() == 'REF':
            return self.counter[self.ref]
        return self.counter[base]

    def __iter__(self):
        """
        Iterate over each base frequency.

        Returns:
            iterator
        """
        return (self[base] for base in self._bases)

    def add_base(self, quality, strand, base):
        """
        Add details for a base at this position.

        Parameters:
            quality (int): The quality of the read
            strand (str): The strand the base is on (+, -, or *)
            base (str): The nucleotide at the position( A, C, G, or T)
        """
        self.qualities.append(quality)
        self.strands.append(strand)
        self.bases.append(base)
        self.counter = False

    def complement(self):
        """Modify all the summarized nucleotides to their complements."""
        self.bases = [self._comp[base] for base in self.bases]
        self.ref = self._comp[self.ref]
        if not self.counter:
            return
        complements = self._comp.items()
        self.counter = {sb: self.counter[bs] for bs, sb in complements}

    def get_variants(self):
        """
        List all detected variants.

        Returns:
            list
        """
        alts = set(self._bases) - {self.ref}
        return [base for base in alts if self[base]]

    def get_strand(self, threshold=0):
        """
        Determine the mean strandedness of a position.

        Parameters:
            threshold (int): Confidence minimum for strand identification

        Returns:
            '+', '-', or '*'
        """
        strand_counts = {'+': 0, '-': 0, '*': 0}
        for idx in self.strands:
            strand_counts[idx] += 1
        total = strand_counts['+'] + strand_counts['-']
        if total == 0:
            return '*'

        strand = max(strand_counts, key=strand_counts.get)
        if strand_counts[strand] / total >= threshold:
            return strand
        return '*'

    def filter_by_strand(self, strand):
        """
        Remove all bases not on the strand.

        Parameters:
            strand (str): Either +, -, or *
        """
        keep = range(len(self.bases))
        keep = [idx for idx in keep if self.strands[idx] == strand]
        self.qualities = self._filter(self.qualities, keep)
        self.strands = self._filter(self.strands, keep)
        self.bases = self._filter(self.bases, keep)
        self.counter = False

    def _filter(self, lst, indx):
        return [lst[idx] for idx in indx]



"""Repeat Sequence Identifier."""
import argparse
import sys

from pysam import FastaFile

from reditools import file_utils


def find_homo_seqs(seq, length=5):
    """
    Locate regions of repeated bases.

    Parameters:
        seq (str): The DNA sequence
        length (int): Minimum number of sequential repeats.

    Yields:
        start, stop, base
    """
    h_base = None
    start = 0
    count = 0

    for pos, base in enumerate(seq):
        if base == h_base:
            count += 1
        else:
            if count >= length:
                yield (start, start + count, h_base)
            count = 0
            start = pos
            h_base = base
    if count >= length:
        yield (start, start + count, h_base)


def parse_options():
    """
    Parse commandline arguments.

    Returns:
        namespace
    """
    parser = argparse.ArgumentParser(
        prog="reditools find-repeats",
        description='REDItools3',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        'file',
        help='The fasta file to be analyzed',
    )
    parser.add_argument(
        '-l',
        '--min-length',
        type=int,
        default=5,
        help='Minimum length of repeat region',
    )
    parser.add_argument(
        '-o',
        '--output',
        default='/dev/stdout',
        help='Destination to write results. Default is to use STDOUT. ' +
        'If the filename ends in .gz, the contents will be gzipped.',
    )

    return parser.parse_args()


def main():
    """Report repetative regions."""
    options = parse_options()
    fasta = FastaFile(options.file)

    if options.output:
        stream = file_utils.open_stream(
            options.output,
            'wt',
            encoding='utf-8',
        )
    else:
        stream = sys.stdout

    for seq_name in fasta.references:
        seq = fasta.fetch(seq_name)
        for region in find_homo_seqs(seq, options.min_length):
            fields = [
                seq_name,
                region[0],
                region[1],
                region[1] - region[0],
                region[2],
            ]
            as_str = [str(_) for _ in fields]
            stream.write('\t'.join(as_str) + '\n')



"""
Analysis system for RNA editing events.

Authors:
    flat - 2017
    ahanden - 2022
"""

from reditools import utils
from reditools.compiled_reads import CompiledReads
from reditools.fasta_file import RTFastaFile
from reditools.logger import Logger
from reditools.rtchecks import RTChecks
from reditools.region_collection import RegionCollection


class RTResult(object):
    """RNA editing analysis for a single base position."""

    def __init__(self, bases, strand, contig, position):
        """
        RNA editing analysis for a single base position.

        Parameters:
            bases (compiledPosition): Bases found by REDItools
            strand (str): Strand of the position
            contig (str): Contig name
            position (int): Genomic position
        """
        self.contig = contig
        self.position = position + 1
        self.bases = bases
        self.strand = strand
        self._variants = bases.get_variants()

    @property
    def variants(self):
        """
        The detected variants at this position.

        Returns:
            list
        """
        ref = self.bases.ref
        return [f'{ref}{base}' for base in self._variants]

    @property
    def mean_quality(self):
        """
        Mean read quality of the base position.

        Returns:
            int
        """
        if self.bases:
            return sum(self.bases.qualities) / len(self.bases)
        return 0

    @property
    def edit_ratio(self):
        """
        Edit ratio.

        Returns:
            float
        """
        if self._variants:
            max_edits = max(self.bases[base] for base in self._variants)
        else:
            max_edits = 0
        return max_edits / (max_edits + self.bases['REF'])

    @property
    def reference(self):
        """
        Base in the reference genome.

        Returns:
            str
        """
        return self.bases.ref

    @property
    def depth(self):
        """
        How many reads cover the position. (post filtering).

        Returns:
            int
        """
        return len(self.bases)

    @property
    def per_base_depth(self):
        """
        How many reads had each base for this position.

        Returns:
            list
        """
        return list(iter(self.bases))


class REDItools(object):
    """Analysis system for RNA editing events."""

    def __init__(self):
        """Create a new REDItools object."""
        self.hostname_string = utils.get_hostname_string()
        self._min_column_length = 1
        self._min_edits = 0
        self._min_edits_per_nucleotide = 0

        self.log_level = Logger.silent_level

        self.strand = 0
        self._use_strand_correction = False
        self.strand_confidence_threshold = 0.5

        self.min_base_quality = 30
        self.min_base_position = 0
        self.max_base_position = float('inf')

        self._rtqc = RTChecks()

        self._min_read_quality = 0

        self._target_regions = RegionCollection()
        self._exclude_regions = RegionCollection()
        self._specific_edits = None

        self.reference = None

        self._include_refs = None

    @property
    def include_refs(self):
        """
        Genome reference bases to report on.

        Returns:
            list
        """
        return self._include_refs

    @property
    def specific_edits(self):
        """
        Specific edit events to report.

        Returns:
            set
        """
        return self._specific_edits

    @specific_edits.setter
    def specific_edits(self, edits):
        if edits == ["ALL"]:
            edits = []
        for alt in edits:
            if not self._verify_alt(alt):
                raise Exception(
                        f'Specific edit "{alt}" is not valid. ' +
                        'Edits must be two character strings of ATCG.')
        self._specific_edits = set(edits)

    def _verify_alt(self, alt):
        if not isinstance(alt, str):
            return False
        if len(alt) != 2:
            return False
        if alt[0] not in 'ATCG' and alt[1] not in 'ATCG':
            return False
        return True

    @property
    def target_regions(self):
        """
        Only report results for these locations.

        Returns:
            list
        """
        return self._target_regions

    def add_target_regions(self, regions):
        """
        Only report results for these locations.

        Parameters:
            regions (iterable): List of Region objects.
        """
        if regions:
            self._target_regions.add_regions(regions)
            self._rtqc.add(self._rtqc.check_target_positions)

    @property
    def log_level(self):
        """
        The logging level.

        Returns:
            Log level
        """
        return self._log_level

    @log_level.setter
    def log_level(self, level):
        """
        Set the class logging level.

        Parameters:
            level (str): logging level
        """
        self._logger = Logger(level)
        self.log = self._logger.log

    @property
    def min_read_quality(self):
        """Minimum read quality for inclusion."""
        return self._min_read_quality  # noqa:DAR201

    @min_read_quality.setter
    def min_read_quality(self, threshold):
        self._min_read_quality = threshold
        function = self._rtqc.check_column_quality
        if self._min_read_quality > 0:
            self._rtqc.add(function)
        else:
            self._rtqc.discard(function)

    @property
    def min_column_length(self):
        """Minimum depth for a position to be reported."""
        return self._min_column_length  # noqa:DAR201

    @min_column_length.setter
    def min_column_length(self, threshold):
        self._min_column_length = threshold
        function = self._rtqc.check_column_min_length
        if threshold > 1:
            self._rtqc.add(function)
        else:
            self._rtqc.discard(function)

    @property
    def min_edits(self):
        """Minimum number of editing events for reporting."""
        return self._min_edits  # noqa:DAR201

    @min_edits.setter
    def min_edits(self, threshold):
        self._min_edits = threshold
        function = self._rtqc.check_column_edit_frequency
        if threshold > 0:
            self._rtqc.add(function)
        else:
            self._rtqc.discard(function)

    @property
    def min_edits_per_nucleotide(self):
        """Minimum number of edits for a single nucleotide for reporting."""
        return self._min_edits_per_nucleotide  # noqa:DAR201

    @min_edits_per_nucleotide.setter
    def min_edits_per_nucleotide(self, threshold):
        self._min_edits_per_nucleotide = threshold
        function = self._rtqc.check_column_min_edits
        if threshold > 0:
            self._rtqc.add(function)
        else:
            self._rtqc.discard(function)

    @property
    def exclude_regions(self):
        """Regions to exclude from analysis"""
        return self._exclude_regions

    def add_exclude_regions(self, regions):
        """
        Regions to exclude from analysis

        Parameters:
            regions (iterable): List of Region objects.
        """
        if regions:
            self._exclude_regions.add_regions(regions)
            self._rtqc.add(self._rtqc.check_exclusions)

    @property
    def max_alts(self):
        """Maximum number of alternative bases for a position."""
        return self._max_alts

    @max_alts.setter
    def max_alts(self, max_alts):
        self._max_alts = max_alts
        function = self._rtqc.check_max_alts
        if max_alts < 3:
            self._rtqc.add(function)
        else:
            self._rtqc.discard(function)

    def exclude(self, regions):
        """
        Explicitly skip specified genomic regions.

        Parameters:
            regions (list): Regions to skip
        """
        for region in regions:
            contig = region.contig
            old_pos = self._exclude_regions.get(contig, set())
            self._exclude_regions[contig] = old_pos | region.enumerate()
        function = self._rtqc.check_exclusions
        if self._exclude_regions:
            self._rtqc.add(function)
        else:
            self._rtqc.discard(function)

    def analyze(self, alignment_manager, region=None):  # noqa:WPS231,WPS213
        """
        Detect RNA editing events.

        Parameters:
            alignment_manager (AlignmentManager): Source of reads
            region (Region): Where to look for edits

        Yields:
            Analysis results for each base position in region
        """
        if region is None:
            region = {}

        # Open the iterator
        self.log(
            Logger.info_level,
            'Fetching data from bams {} [REGION={}]',
            alignment_manager.file_list,
            region,
        )
        read_iter = alignment_manager.fetch_by_position(region=region)
        reads = next(read_iter, None)
        nucleotides = CompiledReads(
            self.strand,
            self.min_base_position,
            self.max_base_position,
            self.min_base_quality,
        )
        if self.reference:
            nucleotides.add_reference(self.reference)
        total = 0
        while reads is not None or not nucleotides.is_empty():
            if nucleotides.is_empty():
                self.log(
                    Logger.debug_level,
                    'Nucleotides is empty: skipping ahead',
                )
                position = alignment_manager.position
                contig = alignment_manager.contig
            else:
                position += 1

            if region.stop and position >= region.stop:
                break
            self.log(
                Logger.debug_level,
                'Analyzing position {} {}',
                contig,
                position,
            )
            # Get all the read(s) starting at position
            if reads and reads[0].reference_start == position:
                self.log(Logger.debug_level, 'Adding {} reads', len(reads))
                total += len(reads)
                nucleotides.add_reads(reads)
                reads = next(read_iter, None)
            # Process edits
            bases = nucleotides.pop(position)
            if not self._rtqc.check(self, bases):
                continue
            column = self._get_column(position, bases, region)
            if column is None:
                self.log(Logger.debug_level, 'Bad column - skipping')
                continue
            if self._specific_edits and \
                    not self._specific_edits & set(column.variants):
                self.log(
                    Logger.debug_level,
                    'Requested edits not found - skipping',
                )
                continue
            self.log(
                Logger.debug_level,
                'Yielding output for {} reads',
                len(bases),
            )
            yield column
        self.log(
            Logger.info_level,
            '[REGION={}] {} total reads',
            region,
            total,
        )

    def use_strand_correction(self):
        """Only reports reads/positions that match `strand`."""
        self._use_strand_correction = True

    def only_one_alt(self):
        """Only report a position if there is less than 2 alt bases."""
        self._rtqc.add(self._rtqc.check_multiple_alts)

    def add_reference(self, reference_fname):
        """
        Use a reference fasta file instead of reference from the BAM files.

        Parameters:
            reference_fname (str): File path to FASTA reference
        """
        self.reference = RTFastaFile(reference_fname)

    def _get_column(self, position, bases, region):
        strand = bases.get_strand(threshold=self.strand_confidence_threshold)
        if self._use_strand_correction:
            bases.filter_by_strand(strand)
            if not bases:
                return None
        if strand == '-':
            bases.complement()

        past_stop = position + 1 >= (region.stop or 0)
        if past_stop or bases is None:
            return None

        return RTResult(bases, strand, region.contig, position)


class REDItoolsDNA(REDItools):
    """
    Analysis system for editing events in DNA.

    Raises:
        ValueError: You cannot set the strand parameter using this class.
    """

    def __init__(self):
        """Create a new REDItoolsDNA object."""
        self.get_position_strand = lambda *_: '*'
        self._get_strand = lambda *_: '*'
        REDItools.__init__(self)

    def set_strand(self, strand):
        """
        Not applicable for DNA analysis.

        Parameters:
            strand (int): N/A

        Raises:
            ValueError: You cannot call this method for DNA analyses.
        """
        raise ValueError('Cannot set strand value if DNA is True')



"""Fast Logging for REDItools."""
import os
import socket
import sys
from datetime import datetime


class Logger(object):
    """Fast logger for REDItools."""

    silent_level = 'SILENT'
    info_level = 'INFO'
    debug_level = 'DEBUG'

    def __init__(self, level):
        """
        Create a new Logger.

        Parameters:
            level (str): either 'INFO' or 'DEBUG'
        """
        hostname = socket.gethostname()
        ip_addr = socket.gethostbyname(hostname)
        pid = os.getpid()
        self.hostname_string = f'{hostname}|{ip_addr}|{pid}'

        if level.upper() == self.debug_level:
            self.log = self._log_all
        elif level.upper() == self.info_level:
            self.log = self._log_info
        else:
            self.log = lambda *_: None

    def _log_all(self, level, message, *args):
        timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        message = message.format(*args)
        sys.stderr.write(
            f'{timestamp} [{self.hostname_string}] ' +
            f'[{level}] {message}\n',
        )

    def _log_info(self, level, message, *args):
        if level == self.info_level:
            self._log_all(level, message, *args)
